import uuid

import aioredis

from catweazle.models.mixins import Format
from catweazle.errors import SessionError


class Sessions(Format):
    def __init__(self, redis_host, redis_port, redis_pass):
        super().__init__()
        self._redis_host = redis_host
        self._redis_port = redis_port
        self._redis_pass = redis_pass
        self._redis = None

    @property
    def redis_host(self):
        return self._redis_host

    @property
    def redis_port(self):
        return self._redis_port

    @property
    def redis_pass(self):
        return self._redis_pass

    @property
    def redis(self):
        return self._redis

    async def init_redis(self):
        self._redis = await aioredis.create_redis_pool(
            (self.redis_host, int(self.redis_port)), password=self.redis_pass, encoding='utf-8')

    async def create(self, user):
        sid = str(uuid.uuid4())
        if not self.redis:
            await self.init_redis()
        await self.redis.set(sid, user)
        await self.redis.expire(sid, 3600)
        return self._format({'id': sid})

    async def delete(self, request):
        sid = request.cookies.get('DEPLOYER_SESSION')
        if not sid:
            return
        try:
            sid = request.cookies.get('DEPLOYER_SESSION')
        except KeyError:
            raise SessionError
        if not self.redis:
            await self.init_redis()
        await self.redis.delete(sid)

    async def get_user(self, request):
        sid = request.cookies.get('DEPLOYER_SESSION')
        if not sid:
            raise SessionError
        if not self.redis:
            await self.init_redis()
        user = await self.redis.get(sid)
        if user is not None:
            await self.redis.expire(sid, 3600)
        return self._format({'id': sid, 'user': user})




import argparse
import asyncio
import configparser
import logging
from logging.handlers import TimedRotatingFileHandler
import os
import sys
import time

from aiohttp import web
import aiotask_context as context
import aiohttp_remotes
import jsonschema
from motor.motor_asyncio import AsyncIOMotorClient
import pymongo

from catweazle.controller.authenticate import Authenticate as ControllerAuthenticate
from catweazle.controller.instances import Instances as ControllerInstances
from catweazle.controller.permissions import Permissions as ControllerPermissions
from catweazle.controller.users import Users as ControllerUsers
from catweazle.controller.users import UsersCredentials as ControllerUsersCredentials

from catweazle.errors import DuplicateResource

from catweazle.middleware import request_id, error_catcher

from catweazle.models.aa import AuthenticationAuthorization as ModelAuthenticationAuthorization
from catweazle.models.aws_route53 import AWSRoute53
from catweazle.models.credentials import Credentials as ModelCredentials
from catweazle.models.foreman_proxy import ForemanProxy as ModelForemanProxy
from catweazle.models.instances import Instances as ModelInstances
from catweazle.models.permissions import Permissions as ModelPermissions
from catweazle.models.sessions import Sessions as ModelSessions
from catweazle.models.users import Users as ModelUsers

from catweazle.schemes.config_main import CHECK_CONFIG_MAIN
from catweazle.schemes.config_main import CHECK_CONFIG_MONGOCOLL, CHECK_CONFIG_MONGOPOOL
from catweazle.schemes.config_main import CHECK_CONFIG_REDISPOOL


def main():
    parser = argparse.ArgumentParser(description="CatWeazle Rest API")

    parser.add_argument("--cfg", dest="cfg", action="store",
                        default="/etc/catweazle/catweazle.ini",
                        help="Full path to configuration")

    subparsers = parser.add_subparsers(help='commands', dest='method')
    subparsers.required = True

    run_parser = subparsers.add_parser('run', help='Start CatWeazle Rest API API')
    run_parser.set_defaults(method='run')

    indicies_parser = subparsers.add_parser('indices', help='create indices and exit')
    indicies_parser.set_defaults(method='indices')

    admin_parser = subparsers.add_parser('create_admin', help='create default admin user')
    admin_parser.set_defaults(method='create_admin')

    parsed_args = parser.parse_args()

    catweazle_restapi = CatWeazleRest(
        cfg=parsed_args.cfg,
    )

    if parsed_args.method == 'run':
        catweazle_restapi.run()

    elif parsed_args.method == 'indices':
        loop = asyncio.get_event_loop()
        loop.run_until_complete(catweazle_restapi.manage_indices())
        loop.close()

    elif parsed_args.method == 'create_admin':
        loop = asyncio.get_event_loop()
        loop.run_until_complete(catweazle_restapi.create_admin())
        loop.close()


class CatWeazleRest:
    def __init__(self, cfg):
        self._config_file = cfg
        self._config = configparser.ConfigParser()
        self._config_dict = None
        self._mongo_pools = dict()
        self._mongo_colls = dict()
        self.log = logging.getLogger('application')
        self.config.read_file(open(self._config_file))
        self._config_dict = self._cfg_to_dict(self.config)
        try:
            jsonschema.validate(self.config_dict['main'], CHECK_CONFIG_MAIN)
        except jsonschema.exceptions.ValidationError as err:
            print("main section: {0}".format(err))
            sys.exit(1)

    def _acc_logging(self):
        acc_handlers = []
        access_log = self.config.get('file:logging', 'acc_log')
        access_retention = self.config.getint('file:logging', 'acc_retention')
        acc_handlers.append(TimedRotatingFileHandler(access_log, 'D', 1, access_retention))

        logger = logging.Logger('aiohttp.access')
        for handler in acc_handlers:
            logger.addHandler(handler)
        return logger

    def _app_logging(self):
        logfmt = logging.Formatter('%(asctime)sUTC - %(levelname)s - %(message)s')
        logfmt.converter = time.gmtime
        app_handlers = []
        aap_level = self.config.get('file:logging', 'app_loglevel')
        app_log = self.config.get('file:logging', 'app_log')
        app_retention = self.config.getint('file:logging', 'app_retention')
        app_handlers.append(TimedRotatingFileHandler(app_log, 'd', 1, app_retention))

        for handler in app_handlers:
            handler.setFormatter(logfmt)
            self.log.addHandler(handler)
        self.log.setLevel(aap_level)
        self.log.debug("application logger is up")

    def run(self):
        self._app_logging()
        self.log.info("starting up")
        loop = asyncio.get_event_loop()
        loop.set_task_factory(context.task_factory)
        self._setup_mongo_pools()
        self._setup_mongo_colls()
        self._validate_redis_pools()

        m_aws_route53 = self._setup_aws_pools()

        m_foreman_proxy, m_foreman_proxy_realm = self._setup_foreman_pools()

        m_instances = ModelInstances(
            self._mongo_colls['instances'],
            domain_suffix=self.config.get('main', 'domain_suffix')
        )
        m_permissions = ModelPermissions(self._mongo_colls['permissions'])
        m_sessions = ModelSessions(
            self.config.get('session:redispool', 'host', fallback='127.0.0.1'),
            self.config.get('session:redispool', 'port', fallback=6379),
            self.config.get('session:redispool', 'pass', fallback=None)
        )
        m_users = ModelUsers(self._mongo_colls['users'])
        m_users_cred = ModelCredentials(self._mongo_colls['users_credentials'])
        m_aa = ModelAuthenticationAuthorization(
            users=m_users,
            users_credentials=m_users_cred,
            permissions=m_permissions,
            sessions=m_sessions
        )

        c_authenticate = ControllerAuthenticate(sessions=m_sessions, users=m_users)
        c_instances = ControllerInstances(
            aa=m_aa, instances=m_instances,
            foreman=m_foreman_proxy, foreman_realm=m_foreman_proxy_realm,
            aws_route53=m_aws_route53,
            indicator_regex=self.config.get('main', 'indicator_regex')
        )
        c_permissions = ControllerPermissions(
            aa=m_aa, permissions=m_permissions
        )
        c_user = ControllerUsers(
            aa=m_aa, credentials=m_users_cred,
            permissions=m_permissions, users=m_users
        )
        c_user_cred = ControllerUsersCredentials(
            aa=m_aa, credentials=m_users_cred, users=m_users
        )

        app = web.Application(middlewares=[request_id, error_catcher, aiohttp_remotes.XForwardedRelaxed().middleware])

        app.router.add_static('/static/', path=str('{0}/static'.format(os.path.dirname(__file__))))

        app.router.add_route('DELETE', '/api/v1/authenticate', c_authenticate.delete)
        app.router.add_route('GET', '/api/v1/authenticate', c_authenticate.get)
        app.router.add_route('POST', '/api/v1/authenticate', c_authenticate.post)

        app.router.add_route('GET', '/api/v1/instances/_search', c_instances.search)
        app.router.add_route('DELETE', '/api/v1/instances/{instance}', c_instances.delete)
        app.router.add_route('GET', '/api/v1/instances/{instance}', c_instances.get)
        app.router.add_route('POST', '/api/v1/instances/{instance}', c_instances.post)

        app.router.add_route('GET', '/api/v1/permissions/_search', c_permissions.search)
        app.router.add_route('DELETE', '/api/v1/permissions/{perm}', c_permissions.delete)
        app.router.add_route('GET', '/api/v1/permissions/{perm}', c_permissions.get)
        app.router.add_route('POST', '/api/v1/permissions/{perm}', c_permissions.post)
        app.router.add_route('PUT', '/api/v1/permissions/{perm}', c_permissions.put)

        app.router.add_route('GET', '/api/v1/users/_search', c_user.search)
        app.router.add_route('DELETE', '/api/v1/users/{user}', c_user.delete)
        app.router.add_route('GET', '/api/v1/users/{user}', c_user.get)
        app.router.add_route('POST', '/api/v1/users/{user}', c_user.post)
        app.router.add_route('PUT', '/api/v1/users/{user}', c_user.put)

        app.router.add_route('GET', '/api/v1/users/{user}/credentials', c_user_cred.get_all)
        app.router.add_route('DELETE', '/api/v1/users/{user}/credentials/{cred}', c_user_cred.delete)
        app.router.add_route('GET', '/api/v1/users/{user}/credentials/{cred}', c_user_cred.get)
        app.router.add_route('POST', '/api/v1/users/{user}/credentials', c_user_cred.post)
        app.router.add_route('PUT', '/api/v1/users/{user}/credentials/{cred}', c_user_cred.put)

        web.run_app(
            app=app,
            host=self.config.get('main', 'host'),
            port=self.config.getint('main', 'port'),
            access_log=self._acc_logging(),
            access_log_format='%a %t "%r" %s %b "%{Referer}i" "%{User-Agent}i" %{X-Request-ID}o'
        )

        self.log.info("shutting down")

    @staticmethod
    def _cfg_to_dict(config):
        result = {}
        for section in config.sections():
            result[section] = {}
            for option in config.options(section):
                try:
                    result[section][option] = config.getint(section, option)
                    continue
                except ValueError:
                    pass
                try:
                    result[section][option] = config.getfloat(section, option)
                    continue
                except ValueError:
                    pass
                try:
                    result[section][option] = config.getboolean(section, option)
                    continue
                except ValueError:
                    pass
                try:
                    result[section][option] = config.get(section, option)
                    continue
                except ValueError:
                    pass
        return result

    def _setup_aws_pools(self):
        self.log.info("setting up aws pools")
        aws = list()
        for section in self.config.sections():
            if section.endswith(':aws'):
                sectionname = section.rsplit(':', 1)[0]
                self.log.debug("setting up aws pool {0}".format(sectionname))
                m_aws_route53 = AWSRoute53(
                    name=sectionname,
                    dry_run=self.config.getboolean('main', 'dry_run', fallback=False),
                    dns_arpa_enable=self.config.getboolean(section, 'dns_arpa_enable'),
                    dns_arpa_zones=self.config.get(section, 'dns_arpa_zones', fallback=None),
                    dns_forward_enable=self.config.getboolean(section, 'dns_forward_enable'),
                    dns_forward_zone=self.config.get(section, 'dns_forward_zone'),
                )
                aws.append(m_aws_route53)
                self.log.debug("setting up aws pool {0} done".format(sectionname))
        self.log.info("setting up aws pools done")
        return aws

    def _setup_foreman_pools(self):
        self.log.info("setting up foreman pools")
        foreman = list()
        foreman_realm = None
        for section in self.config.sections():
            if section.endswith(':foreman'):
                sectionname = section.rsplit(':', 1)[0]
                self.log.debug("setting up foreman pool {0}".format(sectionname))
                m_foreman_proxy = ModelForemanProxy(
                    name=sectionname,
                    dry_run=self.config.getboolean('main', 'dry_run', fallback=False),
                    url=self.config.get(section, 'url'),
                    ssl_crt=self.config.get(section, 'ssl_crt'),
                    ssl_key=self.config.get(section, 'ssl_key'),
                    dns_arpa_enable=self.config.getboolean(section, 'dns_arpa_enable'),
                    dns_arpa_zones=self.config.get(section, 'dns_arpa_zones', fallback=None),
                    dns_forward_enable=self.config.getboolean(section, 'dns_forward_enable'),
                    realm=self.config.get(section, 'realm', fallback=None),
                )
                foreman.append(m_foreman_proxy)
                if m_foreman_proxy.realm:
                    if foreman_realm:
                        self.log.fatal("only one foreman section is allowed to have realm_enable set to true")
                        sys.exit(1)
                    foreman_realm = m_foreman_proxy
                self.log.debug("setting up foreman pool {0} done".format(sectionname))
        self.log.info("setting up foreman pools done")
        return foreman, foreman_realm

    def _setup_mongo_pools(self):
        self.log.info("setting up mongodb connection pools")
        for section in self.config.sections():
            if section.endswith(':mongopool'):
                try:
                    jsonschema.validate(self.config_dict[section], CHECK_CONFIG_MONGOPOOL)
                except jsonschema.exceptions.ValidationError as err:
                    print("{0} section: {1}".format(section, err))
                    sys.exit(1)
                sectionname = section.rsplit(':', 1)[0]
                self.log.debug("setting up mongodb connection pool {0}".format(sectionname))
                pool = AsyncIOMotorClient(
                    host=self.config.get(section, 'hosts'),
                )
                db = pool.get_database(self.config.get(section, 'db'))
                try:
                    user = self.config.get(section, 'user')
                    password = self.config.get(section, 'pass')
                    db.authenticate(user, password)
                    self.log.debug("connection pool {0} is using authentication".format(sectionname))
                except configparser.NoOptionError:
                    self.log.debug("connection pool {0} is not using authentication".format(sectionname))
                self._mongo_pools[sectionname] = db
                self.log.debug("setting up mongodb connection pool {0} done".format(sectionname))
        self.log.info("setting up mongodb connection pools done")

    def _setup_mongo_colls(self):
        self.log.info("setting up mongodb collections")
        for section in self.config.sections():
            if section.endswith(':mongocoll'):
                try:
                    jsonschema.validate(self.config_dict[section], CHECK_CONFIG_MONGOCOLL)
                except jsonschema.exceptions.ValidationError as err:
                    print("{0} section: {1}".format(section, err))
                    sys.exit(1)
                sectionname = section.rsplit(':', 1)[0]
                self.log.debug("setting up mongodb collection {0}".format(sectionname))
                pool_name = self.config.get(section, 'pool')
                coll_name = self.config.get(section, 'coll')
                self.log.debug("mongodb collection {0} is using mongodb connection pool {1}"
                               .format(sectionname, pool_name))
                self.log.debug("mongodb collection {0} is using collection name {1}".format(sectionname, coll_name))
                pool = self._mongo_pools[pool_name]
                coll = pool.get_collection(coll_name)
                self._mongo_colls[sectionname] = coll
                self.log.debug("setting up mongodb collection {0} done".format(sectionname))
        self.log.info("setting up mongodb collections done")

    def _validate_redis_pools(self):
        self.log.info("setting up redis pools")
        for section in self.config.sections():
            if section.endswith(':redispool'):
                try:
                    jsonschema.validate(self.config_dict[section], CHECK_CONFIG_REDISPOOL)
                except jsonschema.exceptions.ValidationError as err:
                    print("{0} section: {1}".format(section, err))
                    sys.exit(1)
        self.log.info("setting up redis pools done")

    @property
    def config(self):
        return self._config

    @property
    def config_dict(self):
        return self._config_dict

    async def create_admin(self):
        self.config.read_file(open(self._config_file))
        self._config_dict = self._cfg_to_dict(self.config)
        self._setup_mongo_pools()
        self._setup_mongo_colls()

        admin = {
            "admin": True,
            "backend": "internal",
            "backend_ref": "default_admin",
            "email": "default_admin@internal",
            "name": "Default Admin User",
            "password": "password"
        }
        try:
            print("creating admin user...")
            m_users = ModelUsers(self._mongo_colls['users'])
            await m_users.create('admin', admin)
            print("done...")
        except DuplicateResource:
            print("admin user already exists...")

    async def manage_indices(self):
        self.config.read_file(open(self._config_file))
        self._config_dict = self._cfg_to_dict(self.config)
        self._setup_mongo_pools()
        self._setup_mongo_colls()

        c_instances = self._mongo_colls["instances"]
        await c_instances.create_index([
            ("id", pymongo.ASCENDING)
        ], unique=True)
        await c_instances.create_index([
            ("fqdn", pymongo.ASCENDING)
        ], unique=True)
        await c_instances.create_index([
            ("ip_address", pymongo.ASCENDING)
        ], unique=True)

        await c_instances.create_index([
            ("dns_indicator", pymongo.ASCENDING),
            ("fqdn", pymongo.ASCENDING),
            ("id", pymongo.ASCENDING),
            ("ip_address", pymongo.ASCENDING)
        ])

        c_permissions = self._mongo_colls["permissions"]
        await c_permissions.create_index([
            ("id", pymongo.ASCENDING)
        ], unique=True)
        await c_permissions.create_index([
            ("permissions", pymongo.ASCENDING)
        ])
        await c_permissions.create_index([
            ("users", pymongo.ASCENDING),
        ])

        c_users = self._mongo_colls["users"]
        await c_users.create_index([
            ("id", pymongo.ASCENDING)
        ], unique=True)

        c_users_credentials = self._mongo_colls["users_credentials"]
        await c_users_credentials.create_index([
            ("owner", pymongo.ASCENDING),
            ("id", pymongo.ASCENDING)
        ], unique=True)



__author__ = 'schlitzer'
import json
import jsonschema
import jsonschema.exceptions

from aiohttp.web import json_response, Response

from catweazle.schemes import AUTHENTICATE_CREATE


class Authenticate:
    def __init__(self, sessions, users):
        self._sessions = sessions
        self._users = users

    @property
    def sessions(self):
        return self._sessions

    @property
    def users(self):
        return self._users

    async def delete(self, request):
        await self.sessions.delete(request)
        response = Response()
        response.set_status(204)
        response.del_cookie('DEPLOYER_SESSION')
        return response

    async def get(self, request):
        result = await self.sessions.get_user(request)
        return json_response(result)

    async def post(self, request):
        payload = await request.json()
        jsonschema.validate(payload, AUTHENTICATE_CREATE, format_checker=jsonschema.draft4_format_checker)
        payload = payload.get('data')
        user = await self.users.check_credentials(payload)
        result = await self.sessions.create(user)
        response = Response(text=json.dumps(result))
        response.content_type = 'application/json'
        response.set_status(201)
        response.set_cookie(
            'DEPLOYER_SESSION', result['data']['id'],
            httponly=True
        )
        return response



__author__ = 'schlitzer'

import datetime
import random
import string
import uuid

from bson.binary import Binary, STANDARD
from passlib.hash import pbkdf2_sha512
import pymongo
import pymongo.errors

from catweazle.models.mixins import ProjectionMixIn
from catweazle.models.mixins import FilterMixIn
from catweazle.models.mixins import Format
from catweazle.models.mixins import ID
from catweazle.errors import CredentialError, MongoConnError, ResourceNotFound


class Credentials(Format, FilterMixIn, ProjectionMixIn, ID):
    def __init__(self, coll):
        super().__init__()
        self.projection_fields = {
            'id': 1,
            'created': 1,
            'description': 1
        }
        self._coll = coll

    @staticmethod
    def _create_secret(token):
        return pbkdf2_sha512.encrypt(str(token), rounds=10, salt_size=32)

    async def check_credential(self, credentials):
        try:
            result = await self._coll.find_one(
                filter={
                    'id': self._str_uuid_2_bin(credentials['id'])
                },
            )
            if not result:
                raise CredentialError
            if not pbkdf2_sha512.verify(credentials['secret'], result['secret']):
                raise CredentialError
            return self._format({'user': result['owner']})
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)

    async def create(self, owner, payload):
        _id = uuid.uuid4()
        secret = ''.join(random.SystemRandom().choice(
            string.ascii_letters + string.digits + '_-.') for _ in range(128))
        created = datetime.datetime.utcnow()
        payload['id'] = Binary(_id.bytes, STANDARD)
        payload['secret'] = self._create_secret(str(secret))
        payload['created'] = created
        payload['owner'] = owner
        try:
            await self._coll.insert_one(payload)
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        result = {
            'id': str(_id),
            'created': str(created),
            'description': payload['description'],
            'secret': str(secret)
        }
        return self._format(result)

    async def delete(self, _id, owner):
        try:
            result = await self._coll.delete_one(
                filter={
                    'id': self._str_uuid_2_bin(_id),
                    'owner': owner
                }
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        if result.deleted_count is 0:
            raise ResourceNotFound(_id)
        return

    async def delete_all_from_owner(self, owner):
        try:
            await self._coll.delete_many(
                filter={"owner": owner}
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)

    async def get(self, _id, owner):
        try:
            result = await self._coll.find_one(
                filter={
                    'id': self._str_uuid_2_bin(_id),
                    'owner': owner}
                ,
                projection=self._projection()
            )
            if result is None:
                raise ResourceNotFound(_id)
            if 'created' in result:
                result['created'] = str(result['created'])
            if 'id' in result:
                result['id'] = str(result['id'])
            return self._format(result)
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)

    async def get_all(self, owner):
        try:
            cursor = self._coll.find(
                filter={'owner': owner},
                projection=self._projection()
            )
            result = list()
            for item in await cursor.to_list(1000):
                if 'created' in item:
                    item['created'] = str(item['created'])
                if 'id' in item:
                    item['id'] = str(item['id'])
                result.append(self._format(item))
            return self._format(result, multi=True)
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)

    async def update(self, _id, owner, payload):
        update = {'$set': {}}
        for k, v in payload.items():
            update['$set'][k] = v
        try:
            result = await self._coll.find_one_and_update(
                filter={
                    'id': self._str_uuid_2_bin(_id),
                    'owner': owner
                },
                update=update,
                projection=self._projection(),
                return_document=pymongo.ReturnDocument.AFTER
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        if result is None:
            raise ResourceNotFound(_id)
        if 'created' in result:
            result['created'] = str(result['created'])
        if 'id' in result:
            result['id'] = str(result['id'])
        return self._format(result)



from uuid import uuid4
import json
import logging

from aiohttp.web import middleware, json_response
import aiotask_context as context
import jsonschema
import pymongo.errors

from catweazle.errors import BaseError
from catweazle.errors import InvalidBody
from catweazle.errors import MongoConnError


log = logging.getLogger('application')


@middleware
async def request_id(request, handler):
    _request_id = request.headers.get('X-Request-ID', None)
    if not _request_id:
        _request_id = str(uuid4())
        request['X-Request-ID'] = _request_id
        context.set('X-Request-ID', _request_id)
    response = await handler(request)
    response.headers['X-Request-ID'] = _request_id
    return response


@middleware
async def error_catcher(request, handler):
    _request_id = request['X-Request-ID']
    try:
        try:
            return await handler(request)
        except (jsonschema.exceptions.ValidationError, json.decoder.JSONDecodeError) as err:
            log.error('{0} received invalid JSON body {1}'.format(_request_id, err))
            raise InvalidBody(err)
        except pymongo.errors.ConnectionFailure as err:
            log.error('{0} error communicating with MongoDB: {1}'.format(_request_id, err))
            raise MongoConnError(err)
    except BaseError as err:
        return json_response(
            data=err.err_rsp,
            status=err.status
        )



import aiohttp
import ssl

import ipaddress

from catweazle.applog import AppLogging
from catweazle.errors import ForemanConnError


class ForemanProxy(object):
    def __init__(
            self, name, dry_run, url, ssl_key, ssl_crt,
            dns_arpa_enable, dns_arpa_zones,
            dns_forward_enable,
            realm
    ):
        self._dns_arpa_enable = dns_arpa_enable
        self._dns_arpa_zones = set()
        for subnet in dns_arpa_zones.split(' '):
            if subnet:
                self._dns_arpa_zones.add(ipaddress.ip_network(subnet))
        self._dns_forward_enable = dns_forward_enable
        self._dry_run = dry_run
        self._name = name
        self._realm = realm
        self._ssl_crt = ssl_crt
        self._ssl_key = ssl_key
        self._url = url
        self.log = AppLogging()

    @property
    def dns_arpa_enable(self):
        return self._dns_arpa_enable

    @property
    def dns_arpa_zones(self):
        return self._dns_arpa_zones

    @property
    def dns_forward_enable(self):
        return self._dns_forward_enable

    @property
    def dry_run(self):
        return self._dry_run

    @property
    def name(self):
        return self._name

    @property
    def realm(self):
        return self._realm

    @property
    def ssl_crt(self):
        return self._ssl_crt

    @property
    def ssl_key(self):
        return self._ssl_key

    @property
    def url(self):
        return self._url

    async def request_delete(self, url):
        if self.dry_run:
            return
        context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        context.load_cert_chain(certfile=self.ssl_crt, keyfile=self.ssl_key)
        async with aiohttp.ClientSession() as session:
            async with await session.delete(url, ssl=context) as resp:
                if resp.status != 200:
                    text = await resp.text()
                    raise ForemanConnError(text)

    async def request_post(self, url, data):
        if self.dry_run:
            return
        context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        context.load_cert_chain(certfile=self.ssl_crt, keyfile=self.ssl_key)
        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=data, ssl=context) as resp:
                if resp.status != 200:
                    text = await resp.text()
                    raise ForemanConnError(text)
                try:
                    data = await resp.json()
                    return data
                except aiohttp.ClientError:
                    pass

    def arpa_responsible(self, ip_address):
        for zone in self.dns_arpa_zones:
            if ip_address in zone:
                return True

    async def create_dns(self, fqdn, ip_address):
        await self.create_arpa_dns(fqdn=fqdn, ip_address=ip_address)
        await self.create_forward_dns(fqdn=fqdn, ip_address=ip_address)

    async def create_arpa_dns(self, fqdn, ip_address):
        self.log.info("{0}:foreman: creating DNS PTR Records for {1} with ip {2}".format(self.name, fqdn, ip_address))
        if not self.dns_arpa_enable:
            self.log.info("{0}:foreman: reverse DNS is disabled".format(self.name))
            return
        url = "{0}{1}".format(self.url, '/dns/')
        body_ptr = aiohttp.FormData()
        ip_addr = ipaddress.IPv4Address(ip_address)
        if self.arpa_responsible(ip_address=ip_addr):
            body_ptr.add_field('fqdn', fqdn)
            body_ptr.add_field('value', ip_addr.reverse_pointer)
            body_ptr.add_field('type', 'PTR')
            await self.request_post(url, body_ptr)
        self.log.info("{0}:foreman: creating DNS PTR Records for {1} with ip {2}, done".format(self.name, fqdn, ip_address))

    async def create_forward_dns(self, fqdn, ip_address):
        self.log.info("{0}:foreman: creating DNS A Records for {1} with ip {2}".format(self.name, fqdn, ip_address))
        if not self.dns_forward_enable:
            self.log.info("{0}:foreman: forward DNS is disabled".format(self.name))
            return
        url = "{0}{1}".format(self.url, '/dns/')
        body_a = aiohttp.FormData()
        body_a.add_field('fqdn', fqdn)
        body_a.add_field('value', ip_address)
        body_a.add_field('type', 'A')
        self.log.info("{0}:foreman: creating DNS A Records for {1} with ip {2}, done".format(self.name, fqdn, ip_address))

        await self.request_post(url, body_a)

    async def delete_dns(self, fqdn, ip_address):
        await self.delete_arpa_dns(fqdn=fqdn, ip_address=ip_address)
        await self.delete_forward_dns(fqdn=fqdn, ip_address=ip_address)

    async def delete_arpa_dns(self, ip_address, fqdn):
        self.log.info("{0}:foreman: deleting DNS PTR Records for {1} with ip {2}".format(self.name, fqdn, ip_address))
        if not self.dns_arpa_enable:
            self.log.info("{0}:foreman: reverse DNS is disabled".format(self.name))
            return
        ip_addr = ipaddress.IPv4Address(ip_address)
        if self.arpa_responsible(ip_address=ip_addr):
            url_ptr = "{0}{1}{2}/PTR".format(self.url, '/dns/', ip_addr.reverse_pointer)
            await self.request_delete(url_ptr)
        self.log.info("{0}:foreman: deleting DNS PTR Records for {1} with ip {2}, done".format(self.name, fqdn, ip_address))

    async def delete_forward_dns(self, ip_address, fqdn):
        self.log.info("{0}:foreman: deleting DNS A Records for {1} with ip {2}".format(self.name, fqdn, ip_address))
        if not self.dns_forward_enable:
            self.log.info("{0}:foreman: forward DNS is disabled".format(self.name))
            return
        url_a = "{0}{1}{2}/A".format(self.url, '/dns/', fqdn)
        await self.request_delete(url_a)
        self.log.info("{0}:foreman: deleting DNS A Records for {1} with ip {2}, done".format(self.name, fqdn, ip_address))

    async def create_realm(self, fqdn):
        self.log.info("{0}:foreman: creating realm entry for {1}".format(self.name, fqdn))
        if self.dry_run:
            return 'dry_run_dummy_pw'
        url = "{0}{1}{2}".format(self.url, '/realm/', self.realm)
        body_a = aiohttp.FormData()
        body_a.add_field('hostname', fqdn)

        data = await self.request_post(url, body_a)
        self.log.info("{0}:foreman: creating realm entry for {1}, done".format(self.name, fqdn))
        return data['randompassword']

    async def delete_realm(self, fqdn):
        self.log.info("{0}:foreman: deleting realm entry for {1}".format(self.name, fqdn))
        url = "{0}{1}{2}/{3}".format(self.url, '/realm/', self.realm, fqdn)
        self.log.info("{0}:foreman: deleting realm entry for {1}, done".format(self.name, fqdn))
        await self.request_delete(url)


#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: proto = await self._create_connection(req, traces, timeout)
#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: File "/opt/catweazle/lib64/python3.6/site-packages/aiohttp/connector.py", line 859, in _create_connection
#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: req, traces, timeout)
#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: File "/opt/catweazle/lib64/python3.6/site-packages/aiohttp/connector.py", line 1004, in _create_direct_connection
#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: raise last_exc
#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: File "/opt/catweazle/lib64/python3.6/site-packages/aiohttp/connector.py", line 986, in _create_direct_connection
#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: req=req, client_error=client_error)
#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: File "/opt/catweazle/lib64/python3.6/site-packages/aiohttp/connector.py", line 943, in _wrap_create_connection
#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: raise client_error(req.connection_key, exc) from exc
#Apr 29 09:50:57 catweazle-3.prod.us-east-1.aws.linux.factset.com catweazle[19156]: aiohttp.client_exceptions.ClientConnectorError: Cannot connect to host fmsmart-dns-aws-1.prod.us-east-1.aws....9', 8443)]
#Hint: Some lines were ellipsized, use -l to show in full.


#!/usr/bin/env python3

import argparse
import logging
import os
import subprocess
import sys
import stat
import time

import requests


def main():
    parser = argparse.ArgumentParser(
        description="Register System based on information provided by CatWeazle. "
                    "Scripts in /etc/catweazle/register.d will be executed in order. "
                    "Scripts will be called with the fqdn as first, and the otp as second argument."
    )

    parser.add_argument("--endpoint", dest="endpoint", action="store", required=True,
                        help="CatWeazle endpoint URL")

    parser.add_argument("--retry", dest="retry", action="store", required=False,
                        default=10, type=int, help="Number of retries for fetching CatWeazle data")

    parser.add_argument("--pre_sleep", dest="pre_sleep", action="store", required=False,
                        default=30, type=int,
                        help="wait specified number of seconds, before doing anything. this might be needed"
                             "because of replication delay between IdM servers.")

    parsed_args = parser.parse_args()

    register = Register(
        endpoint=parsed_args.endpoint,
        retry=parsed_args.retry,
        pre_sleep=parsed_args.pre_sleep,
    )
    register.run()


class Register:
    def __init__(self, endpoint, retry, pre_sleep):
        self.log = logging.getLogger('application')
        self.log.setLevel(logging.DEBUG)
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        self.log.addHandler(handler)

        self._endpoint = endpoint
        self._fqdn = None
        self._instance_id = None
        self._otp = None
        self._pre_sleep = pre_sleep
        self._retry = retry

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def fqdn(self):
        return self._fqdn

    @property
    def instance_id(self):
        if not self._instance_id:
            self._instance_id = requests.get('http://169.254.169.254/latest/meta-data/instance-id').text
        return self._instance_id

    @property
    def otp(self):
        return self._otp

    @property
    def pre_sleep(self):
        return self._pre_sleep

    @property
    def retry(self):
        return self._retry

    def _run_cmd(self, args):
        self.log.info("running command: {0}".format(args))
        p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
        for line in p.stdout:
            self.log.info(line.rstrip())
        p.stdout.close()
        self.log.info("finished running command: {0}".format(args))
        return p.wait()

    def get_cw_data(self):
        self.log.info("Getting CatWeazle Data")
        for _ in range(self.retry):
            self.log.info("Trying to fetch CatWeazle data")
            data = requests.get("{0}/api/v1/instances/{1}".format(self.endpoint, self.instance_id))
            status_code = data.status_code
            data = data.json()
            if status_code is not 200:
                self.log.warning(
                    "Could not fetch instance data, http status was {0}, sleeping for 5 seconds".format(
                        status_code
                    )
                )
            elif 'ipa_otp' in data['data']:
                self.log.info("Success fetching CatWeazle data")
                self._fqdn = data['data']['fqdn']
                self._otp = data['data']['ipa_otp']
                self.log.info("Getting CatWeazle Data, done")
                return
            else:
                self.log.warning(
                    "instance data incomplete, otp token missing, sleeping for 5 seconds")
            time.sleep(5)
        self.log.fatal("instance data could not be fetched, quitting")
        sys.exit(1)

    def get_scripts(self):
        path = '/etc/catweazle/register.d/'
        files = list()
        candidates = os.listdir(path)
        candidates.sort()
        for _file in candidates:
            _file = os.path.join(path, _file)
            self.log.debug("found the file: {0}".format(_file))
            if not os.path.isfile(_file):
                self.log.warning("{0} is not a file".format(_file))
                continue
            if not os.stat(_file).st_uid == 0:
                self.log.warning("file not owned by root")
                continue
            if os.stat(_file).st_mode & stat.S_IXUSR != 64:
                self.log.warning("file not executable by root")
                continue
            if os.stat(_file).st_mode & stat.S_IWOTH == 2:
                self.log.warning("file group writeable")
                continue
            if os.stat(_file).st_mode & stat.S_IWGRP == 16:
                self.log.warning("file world writeable")
                continue
            files.append(_file)
        return files

    def run_scripts(self):
        self.log.info("running registration scripts")
        files = self.get_scripts()
        for _file in files:
            self.log.info("running: {0}".format(_file))
            if self._run_cmd([_file, self.fqdn, self.otp]) != 0:
                self.log.fatal("script failed, stopping!")
                sys.exit(1)
            self.log.info("running: {0} done".format(_file))
        self.log.info("running registration scripts, done")

    def run(self):
        self.log.info("Starting registration process")
        self.log.info("instance-id is {0}".format(self.instance_id))
        self.get_cw_data()
        self.log.info("designated FQDN is {0}".format(self.fqdn))
        self.log.info("sleeping for {0} seconds".format(self.pre_sleep))
        time.sleep(self.pre_sleep)
        self.log.info("sleeping for {0} seconds, done".format(self.pre_sleep))
        self.run_scripts()
        self.log.info("Starting registration process, done")



__author__ = 'schlitzer'

import pymongo
import pymongo.errors

from catweazle.models.mixins import Format, FilterMixIn, PaginationSkipMixIn, ProjectionMixIn, SortMixIn
from catweazle.models.mixins import pagination_from_schema
from catweazle.models.mixins import projection_from_schema
from catweazle.models.mixins import sort_from_schema
from catweazle.errors import MongoConnError, ResourceNotFound, DuplicateResource
from catweazle.schemes import schemes


class Instances(Format, FilterMixIn, PaginationSkipMixIn, ProjectionMixIn, SortMixIn):
    def __init__(self, coll, domain_suffix):
        super().__init__()
        self.pagination_steps = pagination_from_schema(
            schema=schemes, path='/instances/_search'
        )
        self.pagination_limit = self.pagination_steps[-1]
        self.projection_fields = projection_from_schema(
            schema=schemes, path='/instances/_search'
        )
        self.sort_fields = sort_from_schema(
            schema=schemes, path='/instances/_search'
        )
        self._domain_suffix = domain_suffix
        self._coll = coll

    @property
    def domain_suffix(self):
        return self._domain_suffix

    async def _next_num(self, indicator):
        instances = await self.search(dns_indicator=indicator)
        taken = list()
        for instance in instances['data']['results']:
            taken.append(instance['data']['fqdn'])

        fqdn = "{0}{1}".format(indicator, self.domain_suffix)
        for number in range(1, 1000):
            number = str(number)
            if fqdn.replace('NUM', number) not in taken:
                return number

    async def create(self, _id, payload):
        payload['id'] = _id
        fqdn = "{0}{1}".format(payload['dns_indicator'], self.domain_suffix)
        if 'NUM' in payload['dns_indicator']:
            number = await self._next_num(payload['dns_indicator'])
            fqdn = fqdn.replace('NUM', number)
        payload['fqdn'] = fqdn
        try:
            await self._coll.insert_one(payload)
        except pymongo.errors.DuplicateKeyError:
            raise DuplicateResource(_id)
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        return await self.get(_id)

    async def set_ipa_otp(self, _id, ipa_otp):
        if not ipa_otp:
            ipa_otp = ""
        update = {'$set': {}}
        update['$set']['ipa_otp'] = ipa_otp
        try:
            result = await self._coll.find_one_and_update(
                filter={'id': _id},
                update=update,
                projection=self._projection(),
                return_document=pymongo.ReturnDocument.AFTER
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        if result is None:
            raise ResourceNotFound(_id)
        return self._format(result)

    async def delete(self, _id):
        try:
            result = await self._coll.delete_one(filter={
                'id': _id,
            })
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        if result.deleted_count is 0:
            raise ResourceNotFound(_id)
        return

    async def get(self, _id, fields=None):
        try:
            result = await self._coll.find_one(
                filter={
                    'id': _id,
                },
                projection=self._projection(fields)
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        if result is None:
            raise ResourceNotFound(_id)
        return self._format(result)

    async def search(
            self,  instances=None, dns_indicator=None, fqdn=None, ip_address=None,
            fields=None, sort=None, page=None, limit=None):
        query = {}
        self._filter_re(query, 'id', instances)
        self._filter_re(query, 'dns_indicator', dns_indicator)
        self._filter_re(query, 'fqdn', fqdn)
        self._filter_re(query, 'ip_address', ip_address)
        try:
            cursor = self._coll.find(
                filter=query,
                projection=self._projection(fields)
            )
            cursor.sort(self._sort(sort))
            cursor.skip(self._pagination_skip(page, limit))
            cursor.limit(self._pagination_limit(limit))
            result = list()
            for item in await cursor.to_list(self._pagination_limit(limit)):
                result.append(self._format(item))
            return self._format(result, multi=True)
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)



CHECK_CONFIG_MAIN = {
    "type": "object",
    "additionalProperties": False,
    "required": [
        "domain_suffix",
        "host",
        "port",
    ],
    "properties": {
        "domain_suffix": {
            "type": "string",
        },
        "dry_run": {
            "type": "boolean",
        },
        "host": {
            "type": "string",
        },
        "indicator_regex": {
            "type": "string",
        },
        "port": {
            "type": "integer",
            "maximum": 65535,
            "minimum": 1
        },
    }
}

CHECK_CONFIG_MONGOPOOL = {
    "type": "object",
    "additionalProperties": False,
    "required": [
        "hosts",
        "db"
    ],
    "optional": [
        "pass",
        "user"
    ],
    "properties": {
        "hosts": {
            "type": "string",
        },
        "db": {
            "type": "string",
        },
        "pass": {
            "type": "string",
        },
        "user": {
            "type": "string",
        },
    }
}

CHECK_CONFIG_MONGOCOLL = {
    "type": "object",
    "additionalProperties": False,
    "required": [
        "coll",
        "pool"
    ],
    "properties": {
        "coll": {
            "type": "string",
        },
        "pool": {
            "type": "string",
        }
    }
}

CHECK_CONFIG_REDISPOOL = {
    "type": "object",
    "additionalProperties": False,
    "properties": {
        "host": {
            "type": "string",
        },
        "port": {
            "type": "integer",
        },
        "pass": {
            "type": "string",
        },
    }
}




__author__ = 'schlitzer'

__all__ = [
    'AdminError',
    'AlreadyAuthenticatedError',
    'AuthenticationError',
    'BackEndError',
    'BaseError',
    'CredentialError',
    'DuplicateResource',
    'FlowError',
    'ForemanConnError',
    'InvalidBody',
    'InvalidFields',
    'InvalidName',
    'InvalidPaginationLimit',
    'InvalidParameterValue',
    'InvalidSelectors',
    'InvalidSortCriteria',
    'InvalidUUID',
    'ModelError',
    'MongoConnError',
    'PeerReceiverCredentialError',
    'PermError',
    'ResourceNotFound',
    'ResourceInUse',
    'SessionError',
    'SessionCredentialError',
    'StaticPathDisabledError',
    'ValidationError'
]


class BaseError(Exception):
    def __init__(self, status, code, msg):
        super().__init__()
        self.status = status
        self.msg = msg
        self.code = code
        self.err_rsp = {'errors': [{
            "id": self.code,
            "details": self.msg,
            "title": self.msg
        }]}


class AAError(BaseError):
    def __init__(self, status=403, code=1000, msg=None):
        super().__init__(status, code, msg)


class ModelError(BaseError):
    def __init__(self, status=None, code=2000, msg=None):
        super().__init__(status, code, msg)


class ValidationError(BaseError):
    def __init__(self, status=None, code=3000, msg=None):
        super().__init__(status, code, msg)


class FeatureError(BaseError):
    def __init__(self, status=None, code=4000, msg=None):
        super().__init__(status, code, msg)


class BackEndError(BaseError):
    def __init__(self, status=None, code=5000, msg=None):
        super().__init__(status, code, msg)


class AuthenticationError(ModelError):
    def __init__(self):
        super().__init__(
            status=403,
            code=1001,
            msg="Invalid username or Password"
        )


class CredentialError(ModelError):
    def __init__(self):
        super().__init__(
            status=401,
            code=1002,
            msg="Invalid Credentials"
        )


class AlreadyAuthenticatedError(ModelError):
    def __init__(self):
        super().__init__(
            status=403,
            code=1003,
            msg="Already authenticated"
        )


class SessionError(ModelError):
    def __init__(self):
        super().__init__(
            status=403,
            code=1004,
            msg="Invalid or expired session"
        )


class PermError(ModelError):
    def __init__(self, msg):
        super().__init__(
            status=403,
            code=1005,
            msg=msg
        )


class SessionCredentialError(ModelError):
    def __init__(self):
        super().__init__(
            status=403,
            code=1006,
            msg="Neither valid Session or Credentials available"
        )


class AdminError(ModelError):
    def __init__(self):
        super().__init__(
            status=403,
            code=1007,
            msg="Root admin privilege needed for this resource"
        )


class PeerReceiverCredentialError(ModelError):
    def __init__(self):
        super().__init__(
            status=403,
            code=1008,
            msg="Receiver credentials needed for this resource"
        )


class ResourceNotFound(ModelError):
    def __init__(self, resource):
        super().__init__(
            status=404,
            code=2001,
            msg="No resource with ID {0} found".format(resource)
        )


class DuplicateResource(ModelError):
    def __init__(self, resource):
        super().__init__(
            status=400,
            code=2002,
            msg="Duplicate Resource: {0}".format(resource)
        )


class InvalidBody(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3001,
            msg="Invalid post body: {0}".format(err)
        )


class InvalidFields(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3003,
            msg="Invalid field selection: {0}".format(err)
        )


class InvalidSelectors(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3004,
            msg="Invalid selection: {0}".format(err)
        )


class InvalidPaginationLimit(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3005,
            msg="Invalid pagination limit, has to be one of: {0}".format(err)
        )


class InvalidSortCriteria(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3006,
            msg="Invalid sort criteria: {0}".format(err)
        )


class InvalidParameterValue(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3007,
            msg="Invalid parameter value: {0}".format(err)
        )


class InvalidUUID(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3008,
            msg="Invalid uuid: {0}".format(err)
        )


class InvalidName(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3009,
            msg="Invalid Name: {0}".format(err)
        )


class ResourceInUse(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3010,
            msg="Resource is still used: {0}".format(err)
        )


class FlowError(ValidationError):
    def __init__(self, err):
        super().__init__(
            status=400,
            code=3011,
            msg="Flow Error: {0}".format(err)
        )


class StaticPathDisabledError(FeatureError):
    def __init__(self):
        super().__init__(
            status=400,
            code=4002,
            msg="Static path feature is disabled"
        )


class MongoConnError(BackEndError):
    def __init__(self, err):
        super().__init__(
            status=500,
            code=5001,
            msg="MongoDB connection error: {0}".format(err)
        )


class RedisConnError(BackEndError):
    def __init__(self, err):
        super().__init__(
            status=500,
            code=5002,
            msg="Redis connection error: {0}".format(err)
        )


class ForemanConnError(BackEndError):
    def __init__(self, err):
        super().__init__(
            status=500,
            code=5003,
            msg="Foreman connection error: {0}".format(err)
        )



__author__ = 'schlitzer'
import jsonschema
import jsonschema.exceptions

from aiohttp.web import json_response

from catweazle.schemes import PERMISSIONS_CREATE, PERMISSIONS_UPDATE


class Permissions:
    def __init__(self, aa, permissions):
        self._aa = aa
        self._permissions = permissions

    @property
    def aa(self):
        return self._aa

    @property
    def permissions(self):
        return self._permissions

    async def delete(self, request):
        await self.aa.require_admin(request)
        perm = request.match_info['perm']
        await self.permissions.delete_mark(perm)
        return json_response(await self.permissions.delete(perm))

    async def get(self, request):
        await self.aa.require_admin(request)
        perm = request.match_info['perm']
        fields = request.query.get('fields', None)
        return json_response(await self.permissions.get(perm, fields))

    async def post(self, request):
        await self.aa.require_admin(request)
        perm = request.match_info['perm']
        payload = await request.json()
        jsonschema.validate(payload, PERMISSIONS_CREATE, format_checker=jsonschema.draft4_format_checker)
        payload = payload.get('data')
        result = await self.permissions.create(perm, payload)
        return json_response(result, status=201)

    async def put(self, request):
        await self.aa.require_admin(request)
        perm = request.match_info['perm']
        payload = await request.json()
        jsonschema.validate(payload, PERMISSIONS_UPDATE, format_checker=jsonschema.draft4_format_checker)
        payload = payload.get('data')
        result = await self.permissions.update(perm, payload)
        return json_response(result, status=201)

    async def check(self, request):
        await self.aa.require_admin(request)
        result = await self.permissions.check(
            package=request.query.get('package'),
            permission=request.query.get('permission'),
            user=request.query.get('user'),
            fields=request.query.get('fields', None),
        )
        return json_response(result)

    async def search(self, request):
        await self.aa.require_admin(request)
        result = await self.permissions.search(
            permission=request.query.get('permission', None),
            permissions=request.query.get('permissions', None),
            users=request.query.get('users', None),
            fields=request.query.get('fields', None),
            sort=request.query.get('sort', None),
            page=request.query.get('page', None),
            limit=request.query.get('limit', None)
        )
        return json_response(result)



__author__ = 'schlitzer'
import jsonschema
import jsonschema.exceptions
import re

from aiohttp.web import json_response

from catweazle.applog import AppLogging
from catweazle.schemes import INSTANCES_CREATE
from catweazle.errors import BackEndError
from catweazle.errors import BaseError
from catweazle.errors import InvalidName
from catweazle.errors import ForemanConnError
from catweazle.errors import ModelError
from catweazle.errors import ResourceNotFound


class Instances:
    def __init__(self, aa, aws_route53, foreman, foreman_realm, instances, indicator_regex):
        self._aa = aa
        self._aws_route53 = aws_route53
        self._foreman = foreman
        self._foreman_realm = foreman_realm
        self._instances = instances
        self._indicator_regex = re.compile(indicator_regex)
        self.log = AppLogging()

    @property
    def aa(self):
        return self._aa

    @property
    def aws_route53(self):
        return self._aws_route53

    @property
    def foreman(self):
        return self._foreman

    @property
    def foreman_realm(self):
        return self._foreman_realm

    @property
    def indicator_regex(self):
        return self._indicator_regex

    @property
    def instances(self):
        return self._instances

    async def delete(self, request):
        instance = request.match_info['instance']
        await self.aa.require(request, 'INSTANCE:DELETE')
        _instance = await self.instances.get(instance)
        for foreman in self.foreman:
            try:
                await foreman.delete_dns(
                    fqdn=_instance['data']['fqdn'],
                    ip_address=_instance['data']['ip_address']
                )
            except ForemanConnError as err:
                self.log.error(err.msg)
        for aws in self.aws_route53:
            try:
                await aws.delete(
                    fqdn=_instance['data']['fqdn'],
                    ip_address=_instance['data']['ip_address']
                )
            except ForemanConnError as err:
                self.log.error(err.msg)
        try:
            if self.foreman_realm:
                await self.foreman_realm.delete_realm(
                    fqdn=_instance['data']['fqdn'],
                )
        except ForemanConnError as err:
            self.log.error(err.msg)
        result = await self.instances.delete(instance)
        return json_response(result)

    async def get(self, request):
        instance = request.match_info['instance']
        try:
            await self.aa.require(request, instance)
        except ModelError as err:
            try:
                _result = await self.instances.get(instance, 'id,ip_address')
                self.log.info("remote ip is: {0}".format(request.remote))
                self.log.info(request.headers)
                if request.remote != _result['data']['ip_address']:
                    raise err
            except ResourceNotFound:
                raise err

        fields = request.query.get('fields', None)
        result = await self.instances.get(instance, fields)
        return json_response(result)

    async def post(self, request):
        instance = request.match_info['instance']
        await self.aa.require(request, 'INSTANCE:POST')
        payload = await request.json()
        jsonschema.validate(payload, INSTANCES_CREATE, format_checker=jsonschema.draft4_format_checker)
        payload = payload.get('data')
        if not self.indicator_regex.match(instance):
            raise InvalidName('{0}'.format(instance))
        result = await self.instances.create(instance, payload)
        try:
            for foreman in self.foreman:
                await foreman.create_dns(
                    fqdn=result['data']['fqdn'],
                    ip_address=result['data']['ip_address']
                )
            for aws in self.aws_route53:
                await aws.create(
                    fqdn=result['data']['fqdn'],
                    ip_address=result['data']['ip_address']
                )
            if self.foreman_realm:
                ipa_otp = await self.foreman_realm.create_realm(
                    fqdn=result['data']['fqdn'],
                )
                result = await self.instances.set_ipa_otp(instance, ipa_otp)
        except BaseError as err:
            self.log.error("something went wrong, rolling back")
            self.log.error(err.msg)
            await self.delete(request)
            raise BackEndError('Could not create instance in backend, please check logs')
        return json_response(result, status=201)

    async def search(self, request):
        result = await self.instances.search(
            instances=request.query.get('instances', None),
            dns_indicator=request.query.get('dns_indicator', None),
            ip_address=request.query.get('ip_address', None),
            fqdn=request.query.get('fqdn', None),
            fields=request.query.get('fields', None),
            sort=request.query.get('sort', None),
            page=request.query.get('page', None),
            limit=request.query.get('limit', None),
        )
        return json_response(result)



import logging

import aiotask_context as context


class AppLogging:
    def __init__(self):
        self.log = logging.getLogger('application')
        self.context = context

    def info(self, msg):
        self.log.info('{0} {1}'.format(self.context.get('X-Request-ID'), msg))

    def warning(self, msg):
        self.log.warning('{0} {1}'.format(self.context.get('X-Request-ID'), msg))

    def error(self, msg):
        self.log.error('{0} {1}'.format(self.context.get('X-Request-ID'), msg))

    def critical(self, msg):
        self.log.critical('{0} {1}'.format(self.context.get('X-Request-ID'), msg))

    def fatal(self, msg):
        self.log.fatal('{0} {1}'.format(self.context.get('X-Request-ID'), msg))

    def debug(self, msg):
        self.log.debug('{0} {1}'.format(self.context.get('X-Request-ID'), msg))



import os

from yaml import load, Loader

with open('{0}/../static/swagger.yaml'.format(os.path.dirname(__file__)), 'r') as scheme_source:
    schemes = load(scheme_source, Loader=Loader)

AUTHENTICATE_CREATE = schemes['components']['schemas']['Authenticate_POST']

CREDENTIALS_CREATE = schemes['components']['schemas']['Common_Credential_POST']

INSTANCES_CREATE = schemes['components']['schemas']['Instance_POST']

PERMISSIONS_CREATE = schemes['components']['schemas']['Permission_POST']
PERMISSIONS_UPDATE = schemes['components']['schemas']['Permission_PUT']

USERS_CREATE = schemes['components']['schemas']['User_POST']
USERS_UPDATE = schemes['components']['schemas']['User_PUT']



from setuptools import setup, find_packages

setup(
    name='CatWeazle',
    version='0.0.16',
    description='CatWeazle, dynamic DNS and Redhat IdM/FreeIPA registration system',
    long_description="""
dynamically create and delete DNS records for volatile linux systems, as well as redhat IDM de/registration.

Copyright (c) 2019, Stephan Schultchen.

License: MIT (see LICENSE for details)
    """,
    packages=find_packages(),
    scripts=[
        'contrib/catweazle',
        'contrib/catweazle_register',
    ],
    url='https://github.com/schlitzered/CatWeazle',
    license='MIT',
    author='schlitzer',
    author_email='stephan.schultchen@gmail.com',
    include_package_data=True,
    test_suite='test',
    platforms='posix',
    classifiers=[
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3'
    ],
    install_requires=[
        "aioboto3",
        "aiohttp",
        "aiotask-context",
        "aiohttp-remotes",
        "ipaddress",
        "jsonschema",
        "motor",
        "pyyaml",
        "passlib",
        "requests",
        "aioredis",
    ],
    keywords=[
        'freeipa', 'redhat idm', 'dns'
    ]
)



import ipaddress

import aioboto3

from botocore.exceptions import ClientError

from catweazle.applog import AppLogging

# exception if record not found, and trying to delete
# botocore.errorfactory.InvalidChangeBatch:


class AWSRoute53(object):
    def __init__(
            self, name, dry_run,
            dns_arpa_enable, dns_arpa_zones,
            dns_forward_enable, dns_forward_zone,
    ):
        self._dns_arpa_enable = dns_arpa_enable
        self._dns_arpa_zones = dict()
        for subnet in dns_arpa_zones.split(' '):
            subnet, subnet_id = subnet.rsplit(':')
            self._dns_arpa_zones[ipaddress.ip_network(subnet)] = subnet_id
        self._dns_forward_enable = dns_forward_enable
        self._dns_forward_zone = dns_forward_zone
        self._name = name
        self._dry_run = dry_run
        self.log = AppLogging()

    @property
    def dns_arpa_enable(self):
        return self._dns_forward_enable

    @property
    def dns_arpa_zones(self):
        return self._dns_arpa_zones

    @property
    def dns_forward_enable(self):
        return self._dns_forward_enable

    @property
    def dns_forward_zone(self):
        return self._dns_forward_zone

    @property
    def dry_run(self):
        return self._dry_run

    @property
    def name(self):
        return self._name

    def arpa_responsible(self, ip_address):
        for subnet, subnet_id in self.dns_arpa_zones.items():
            if ip_address in subnet:
                self.log.info("{0}:aws: responsible DNS zone {1} with id {2}".format(self.name, subnet, subnet_id))
                return subnet_id

    async def create(self, fqdn, ip_address):
        self.log.info("{0}:aws: creating DNS Records for {1} with ip {2}".format(self.name, fqdn, ip_address))
        await self.create_arpa(fqdn=fqdn, ip_address=ip_address)
        await self.create_forward(fqdn=fqdn, ip_address=ip_address)
        self.log.info("{0}:aws: creating DNS Records for {1} with ip {2}, done".format(self.name, fqdn, ip_address))

    async def create_forward(self, fqdn, ip_address):
        if not self.dns_forward_enable:
            self.log.info("{0}:aws: forward DNS is disabled".format(self.name))
            return
        if self.dry_run:
            return
        async with aioboto3.client('route53') as client:
            try:
                await client.change_resource_record_sets(
                    HostedZoneId=self.dns_forward_zone,
                    ChangeBatch={
                        "Comment": "CatWeazle: add {0} in A {1}".format(fqdn, ip_address),
                        "Changes": [
                            {
                                "Action": "UPSERT",
                                "ResourceRecordSet": {
                                    "Name": fqdn,
                                    "Type": "A",
                                    "TTL": 300,
                                    "ResourceRecords": [{"Value": ip_address}]
                                }
                            }]
                    })
            except ClientError:
                pass

    async def create_arpa(self, fqdn, ip_address):
        if not self.dns_arpa_enable:
            self.log.info("{0}:aws: reverse DNS is disabled".format(self.name))
            return
        ip_addr = ipaddress.IPv4Address(ip_address)
        subnet_id = self.arpa_responsible(ip_address=ip_addr)
        if not subnet_id:
            return
        if self.dry_run:
            return
        async with aioboto3.client('route53') as client:
            try:
                await client.change_resource_record_sets(
                    HostedZoneId=subnet_id,
                    ChangeBatch={
                        "Comment": "CatWeazle: add {0} in PTR {1}".format(fqdn, ip_address),
                        "Changes": [
                            {
                                "Action": "UPSERT",
                                "ResourceRecordSet": {
                                    "Name": ip_addr.reverse_pointer,
                                    "Type": "PTR",
                                    "TTL": 300,
                                    "ResourceRecords": [{"Value": fqdn}]
                                }
                            }]
                    })
            except ClientError:
                pass

    async def delete(self, fqdn, ip_address):
        self.log.info("{0}:aws: deleting DNS Records for {1} with ip {2}".format(self.name, fqdn, ip_address))
        await self.delete_arpa(fqdn=fqdn, ip_address=ip_address)
        await self.delete_forward(fqdn=fqdn, ip_address=ip_address)
        self.log.info("{0}:aws: deleting DNS Records for {1} with ip {2}, done".format(self.name, fqdn, ip_address))

    async def delete_forward(self, fqdn, ip_address):
        if not self.dns_arpa_enable:
            self.log.info("{0}:aws: forward DNS is disabled".format(self.name))
            return
        if self.dry_run:
            return
        async with aioboto3.client('route53') as client:
            try:
                await client.change_resource_record_sets(
                    HostedZoneId=self.dns_forward_zone,
                    ChangeBatch={
                        "Comment": "CatWeazle: remove {0} in A {1}".format(fqdn, ip_address),
                        "Changes": [
                            {
                                "Action": "DELETE",
                                "ResourceRecordSet": {
                                    "Name": fqdn,
                                    "Type": "A",
                                    "TTL": 300,
                                    "ResourceRecords": [{"Value": ip_address}]
                                }
                            }]
                    })
            except ClientError:
                pass

    async def delete_arpa(self, fqdn, ip_address):
        if not self.dns_arpa_enable:
            self.log.info("{0}:aws: reverse DNS is disabled".format(self.name))
            return
        ip_addr = ipaddress.IPv4Address(ip_address)
        subnet_id = self.arpa_responsible(ip_address=ip_addr)
        if not subnet_id:
            return
        if self.dry_run:
            return
        async with aioboto3.client('route53') as client:
            try:
                await client.change_resource_record_sets(
                    HostedZoneId=subnet_id,
                    ChangeBatch={
                        "Comment": "CatWeazle: remove {0} in PTR {1}".format(fqdn, ip_address),
                        "Changes": [
                            {
                                "Action": "DELETE",
                                "ResourceRecordSet": {
                                    "Name": ip_addr.reverse_pointer,
                                    "Type": "PTR",
                                    "TTL": 300,
                                    "ResourceRecords": [{"Value": fqdn}]
                                }
                            }]
                    })
            except ClientError:
                pass



__author__ = 'schlitzer'

from catweazle.errors import PermError, CredentialError, SessionCredentialError, AdminError
from catweazle.errors import SessionError


class AuthenticationAuthorization(object):
    def __init__(self, users, users_credentials, permissions, sessions):
        self.users = users
        self.users_credentials = users_credentials
        self.sessions = sessions
        self.permissions = permissions

    async def get_user_from_session(self, request):
        result = await self.sessions.get_user(request)
        return result['data']['user']

    @staticmethod
    async def get_user_credential(request):
        result = {}
        _id = request.headers.get('X-ID', False)
        secret = request.headers.get('X-SECRET', False)
        if _id and secret:
            result['id'] = _id
            result['secret'] = secret
            return result
        raise CredentialError

    async def get_user_from_credential(self, request):
        credential = await self.get_user_credential(request)
        return await self.users_credentials.check_credential(credential)

    async def get_user(self, request):
        try:
            user = await self.get_user_from_session(request)
        except SessionError:
            try:
                user = await self.get_user_from_credential(request)
                user = user['data']['user']
            except CredentialError:
                raise SessionCredentialError
        return user

    async def require_admin(self, request):
        _is_admin = await self.users.is_admin(await self.get_user(request))
        if not _is_admin:
            raise AdminError
        return True

    async def require(self, request, permission):
        user = await self.get_user(request)
        try:
            return await self.require_admin(request)
        except AdminError:
            try:
                await self.permissions.check(
                    user=user,
                    permission=permission,
                    fields='id'
                )
                return True
            except PermError as err:
                raise err




__author__ = 'schlitzer'

import uuid

import pymongo
from bson.binary import Binary, STANDARD

from catweazle.errors import InvalidSelectors
from catweazle.errors import InvalidFields
from catweazle.errors import InvalidParameterValue
from catweazle.errors import InvalidPaginationLimit
from catweazle.errors import InvalidSortCriteria
from catweazle.errors import InvalidUUID


def pagination_from_schema(schema, path):
    for item in schema['paths'][path]['get']['parameters']:
        if item['name'] == 'limit':
            fields = (item['description'].split(':', maxsplit=1)[1].split(','))
            fields = [x.strip(' ') for x in fields]
            result = list()
            for field in fields:
                result.append(int(field))
            result.sort()
            return result


def projection_from_schema(schema, path):
    for item in schema['paths'][path]['get']['parameters']:
        if item['name'] == 'fields':
            fields = (item['description'].split(':', maxsplit=1)[1].split(','))
            fields = [x.strip(' ') for x in fields]
            result = dict()
            for field in fields:
                result[field] = 1
            return result


def sort_from_schema(schema, path):
    for item in schema['paths'][path]['get']['parameters']:
        if item['name'] == 'sort':
            fields = (item['description'].split(':', maxsplit=1)[1].split(','))
            fields = [x.strip(' ') for x in fields]
            result = list()
            for field in fields:
                result.append([field, pymongo.ASCENDING])
            return result


class ID(object):
    @staticmethod
    def _str_uuid_2_bin(_id):
        try:
            return Binary(uuid.UUID(_id).bytes, STANDARD)
        except ValueError:
            raise InvalidUUID(_id)

    def _str_uuid_2_bin_list(self, _ids):
        uuids = list()
        for _id in _ids:
            try:
                uuids.append(self._str_uuid_2_bin(_id))
            except ValueError:
                raise InvalidUUID(_id)
        return uuids

    @staticmethod
    def _bin_uuid_2_str_list(_ids):
        uuids = list()
        for _id in _ids:
            try:
                uuids.append(str(_id))
            except ValueError:
                raise InvalidUUID(_id)
        return uuids


class Format(object):
    @staticmethod
    def _format(item, multi=False, keep_id=False):
        if multi:
            return {"data": {"results": item}}
        else:
            if not keep_id:
                item.pop('_id', None)
        return {"data": item}


class FilterMixIn(object):
    @staticmethod
    def _filter_boolean(query, field, selector):
        if selector is None:
            return
        if selector in [True, 'true', 'True', '1']:
            selector = True
        elif selector in [False, 'false', 'False', '0']:
            selector = False
        else:
            raise InvalidSelectors('Selector is not a boolean')
        query[field] = selector

    @staticmethod
    def _filter_list(query, field, selector):
        if selector is None:
            return
        if type(selector) is not list:
            selector = list(set(selector.split(',')))
        query[field] = {'$in': selector}

    @staticmethod
    def _filter_list_uuid(query, field, selector):
        if selector is None:
            return
        if type(selector) is not list:
            selector = list(set(selector.split(',')))
        uuid_selector = list()
        for _id in selector:
            uuid_selector.append(Binary(uuid.UUID(_id).bytes, STANDARD))
        query[field] = {'$in': uuid_selector}

    @staticmethod
    def _filter_re(query, field, selector):
        if selector is None:
            return
        query[field] = {'$regex': selector}


class ProjectionMixIn(object):
    def __init__(self):
        self.projection_fields = {}

    def _projection(self, fields=None):
        if not fields:
            return self.projection_fields
        fields = fields.split(sep=',')
        for field in fields:
            if field not in self.projection_fields:
                raise InvalidFields('{0} is not a valid field'.format(field))
        result = {}
        for field in fields:
            result[field] = 1
        return result


class PaginationSkipMixIn(object):
    pagination_limit = 1000
    pagination_steps = [10, 25, 50, 100, 250, 500, 1000]

    def _pagination_skip(self, page=None, limit=None):
        if not page:
            page = 0
        else:
            try:
                page = int(page)
                if page < 1:
                    raise ValueError
            except ValueError:
                raise InvalidParameterValue("got {0} expected integer for parameter page".format(page))
        if not limit:
            limit = self.pagination_limit
        else:
            try:
                limit = int(limit)
                if limit < 1:
                    raise ValueError
            except ValueError:
                raise InvalidPaginationLimit(self.pagination_steps)
        return page * limit

    def _pagination_limit(self, limit=None):
        if not limit:
            limit = self.pagination_limit
        else:
            limit = int(limit)
            if limit not in self.pagination_steps:
                raise InvalidPaginationLimit(self.pagination_steps)
        return limit


class SortMixIn(object):
    sort_fields = []

    def _sort(self, sort=None, strict_order=True):
        if not sort:
            return self.sort_fields
        result = []
        items = []
        sort = sort.split(sep=',')
        for item in sort:
            if item.startswith('-'):
                order = pymongo.DESCENDING
                item = item[1:]
            else:
                order = pymongo.ASCENDING
            self._sort_valid_criteria(item)
            if item not in items:
                items.append(item)
            else:
                raise InvalidSortCriteria('{0} has been specified multiple times'.format(item))
            result.append((item, order))
        if strict_order:
            result = self._sort_strict_order(result)
        return result

    def _sort_valid_criteria(self, item):
        for allowed in self.sort_fields:
            if allowed[0] == item:
                return
        raise InvalidSortCriteria('{0} is not allowed as sort criteria'.format(item))

    def _sort_strict_order(self, items):
        for element in range(len(self.sort_fields)):
            try:
                if not self.sort_fields[element][0] == items[element][0]:
                    raise InvalidSortCriteria('sort criteria number {0} should be {1} but is {2}'.format(
                        element, self.sort_fields[element][0], items[element][0]
                    ))
            except IndexError:
                items.append(self.sort_fields[element])
        return items



__author__ = 'schlitzer'

import pymongo
import pymongo.errors

from catweazle.models.mixins import Format, FilterMixIn, PaginationSkipMixIn, ProjectionMixIn, SortMixIn
from catweazle.models.mixins import pagination_from_schema
from catweazle.models.mixins import projection_from_schema
from catweazle.models.mixins import sort_from_schema
from catweazle.errors import MongoConnError, PermError, ResourceNotFound, DuplicateResource
from catweazle.schemes import schemes


class Permissions(Format, FilterMixIn, PaginationSkipMixIn, ProjectionMixIn, SortMixIn):
    def __init__(self, coll):
        super().__init__()
        self.pagination_steps = pagination_from_schema(
            schema=schemes, path='/permissions/_search'
        )
        self.pagination_limit = self.pagination_steps[-1]
        self.projection_fields = projection_from_schema(
            schema=schemes, path='/permissions/_search'
        )
        self.sort_fields = sort_from_schema(
            schema=schemes, path='/permissions/_search'
        )
        self._coll = coll

    async def check(self, user, fields, permission):
        query = {
            'deleting': False,
            'users': user
        }
        self._filter_list(query, 'permissions', permission)
        try:
            result = await self._coll.find_one(
                filter=query,
                projection=self._projection(fields)
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        if result is None:
            raise PermError(
                msg="Required Permissions {0} not found for user {1}".format(
                    permission, user
                )
            )
        return result

    async def create(self, _id, payload):
        payload['id'] = _id
        payload['deleting'] = False
        try:
            await self._coll.insert_one(payload)
        except pymongo.errors.DuplicateKeyError:
            raise DuplicateResource(_id)
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        return await self.get(_id)

    async def delete(self, _id):
        try:
            result = await self._coll.delete_one(filter={
                'id': _id,
            })
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        if result.deleted_count is 0:
            raise ResourceNotFound(_id)
        return

    async def delete_mark(self, _id):
        update = {'$set': {'deleting': True}}
        try:
            await self._coll.update_one(
                filter={'id': _id},
                update=update,
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)

    async def get(self, _id, fields=None):
        try:
            result = await self._coll.find_one(
                filter={
                    'id': _id,
                    'deleting': False
                },
                projection=self._projection(fields)
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        if result is None:
            raise ResourceNotFound(_id)
        return self._format(result)

    async def delete_user_from_all(self, user):
        try:
            await self._coll.update_many(
                filter={"users": user},
                update={"$pull": {"users": user}}
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)

    async def search(
            self, permission, permissions=None, users=None,
            fields=None, sort=None, page=None, limit=None):
        query = {}
        self._filter_re(query, 'id', permission)
        self._filter_re(query, 'permissions', permissions)
        self._filter_re(query, 'users', users)
        self._filter_boolean(query, 'deleting', False)
        try:
            cursor = self._coll.find(
                filter=query,
                projection=self._projection(fields)
            )
            cursor.sort(self._sort(sort))
            cursor.skip(self._pagination_skip(page, limit))
            cursor.limit(self._pagination_limit(limit))
            result = list()
            for item in await cursor.to_list(self._pagination_limit(limit)):
                result.append(self._format(item))
            return self._format(result, multi=True)
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)

    async def update(self, _id, payload):
        update = {'$set': {}}
        for k, v in payload.items():
            update['$set'][k] = v
        try:
            result = await self._coll.find_one_and_update(
                filter={'id': _id},
                update=update,
                projection=self._projection(),
                return_document=pymongo.ReturnDocument.AFTER
            )
        except pymongo.errors.ConnectionFailure as err:
            raise MongoConnError(err)
        if result is None:
            raise ResourceNotFound(_id)
        return self._format(result)



__author__ = 'schlitzer'
from aiohttp.web import json_response
import jsonschema
import jsonschema.exceptions

from catweazle.errors import ResourceNotFound
from catweazle.schemes import CREDENTIALS_CREATE
from catweazle.schemes import USERS_CREATE, USERS_UPDATE


class Users:
    def __init__(self, aa, credentials, permissions, users):
        self._aa = aa
        self._credentials = credentials
        self._permissions = permissions
        self._users = users

    @property
    def aa(self):
        return self._aa

    @property
    def credentials(self):
        return self._credentials

    @property
    def permissions(self):
        return self._permissions

    @property
    def users(self):
        return self._users

    async def delete(self, request):
        user = request.match_info['user']
        if user == '_self':
            user = await self.aa.get_user(request)
        else:
            await self.aa.require_admin(request)
            await self.users.delete_mark(user)
            await self.permissions.delete_user_from_all(user)
            await self.credentials.delete_all_from_owner(user)
        return json_response(await self.users.delete(user))

    async def get(self, request):
        user = request.match_info['user']
        if user == '_self':
            user = await self.aa.get_user(request)
        else:
            await self.aa.require_admin(request)
        fields = request.query.get('fields', None)
        return json_response(await self.users.get(user, fields))

    async def post(self, request):
        user = request.match_info['user']
        await self.aa.require_admin(request)
        payload = await request.json()
        jsonschema.validate(payload, USERS_CREATE, format_checker=jsonschema.draft4_format_checker)
        payload = payload.get('data')
        if 'admin' not in payload:
            payload['admin'] = False
        payload['backend'] = 'internal'
        payload['backend_ref'] = user
        result = await self.users.create(user, payload)
        return json_response(result, status=201)

    async def put(self, request):
        user = request.match_info['user']
        payload = await request.json()
        jsonschema.validate(payload, USERS_UPDATE, format_checker=jsonschema.draft4_format_checker)
        payload = payload.get('data')
        if user == '_self':
            payload.pop('admin', None)
            user = await self.aa.get_user(request)
        else:
            await self.aa.require_admin(request)
        return json_response(await self.users.update(user, payload))

    async def search(self, request):
        await self.aa.require_admin(request)
        result = await self.users.search(
            _id=request.query.get('id', None),
            fields=request.query.get('fields', None),
            sort=request.query.get('sort', None),
            page=request.query.get('page', None),
            limit=request.query.get('limit', None)
        )
        return json_response(result)


class UsersCredentials:
    def __init__(self, aa, users, credentials):
        self._aa = aa
        self._credentials = credentials
        self._users = users

    @property
    def aa(self):
        return self._aa

    @property
    def credentials(self):
        return self._credentials

    @property
    def users(self):
        return self._users

    async def delete(self, request):
        user = request.match_info['user']
        cred = request.match_info['cred']
        if user == '_self':
            user = await self.aa.get_user(request)
        else:
            await self.aa.require_admin(request)
        if not await self.users.resource_exists(user):
            raise ResourceNotFound(user)
        return json_response(await self.credentials.delete(cred, user))

    async def get(self, request):
        user = request.match_info['user']
        cred = request.match_info['cred']
        if user == '_self':
            user = await self.aa.get_user(request)
        else:
            await self.aa.require_admin(request)
        if not await self.users.resource_exists(user):
            raise ResourceNotFound(user)
        return json_response(await self.credentials.get(cred, user))

    async def get_all(self, request):
        user = request.match_info['user']
        if user == '_self':
            user = await self.aa.get_user(request)
        else:
            await self.aa.require_admin(request)
        if not await self.users.resource_exists(user):
            raise ResourceNotFound(user)
        return json_response(await self.credentials.get_all(user))

    async def post(self, request):
        user = request.match_info['user']
        if user == '_self':
            user = await self.aa.get_user(request)
        else:
            await self.aa.require_admin(request)
        payload = await request.json()
        jsonschema.validate(payload, CREDENTIALS_CREATE, format_checker=jsonschema.draft4_format_checker)
        payload = payload.get('data')
        if not await self.users.resource_exists(user):
            raise ResourceNotFound(user)
        return json_response(await self.credentials.create(user, payload), status=201)

    async def put(self, request):
        user = request.match_info['user']
        cred = request.match_info['cred']
        if user == '_self':
            user = await self.aa.get_user(request)
        else:
            await self.aa.require_admin(request)
        payload = await request.json()
        jsonschema.validate(payload, CREDENTIALS_CREATE, format_checker=jsonschema.draft4_format_checker)
        payload = payload.get('data')
        if not await self.users.resource_exists(user):
            raise ResourceNotFound(user)
        return json_response(await self.credentials.update(cred, user, payload), status=201)
